/**
 * JASMINe
 * Copyright (C) 2010 Bull S.A.S.
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * $Id: PurgeDataBase.java 7850 2011-03-21 10:45:17Z danesa $
 */
package org.ow2.jasmine.monitoring.commands;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.text.Format;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.commons.cli.AlreadySelectedException;
import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;

import org.ow2.easybeans.component.smartclient.spi.SmartContextFactory;

import org.ow2.jasmine.monitoring.eventswitch.beans.JasmineEventPurgeTaskSLBRemote;
import org.ow2.jasmine.event.messages.JasmineEventPurgeTask;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Implements 'purge' command, a command that allows JASMINe DB purge.
 * It is based on a StalessSessionBean deployed on the JASMINe's host server.
 * @author danesa
 *
 */
public class PurgeDataBase {

    private Log logger = LogFactory.getLog(this.getClass());

    /**
     * JNDI name of the StalessSessionBean that implements Purge function.
     */
    private final static String EJB_JNDI_NAME = "db-ejb/purgetask";

    /**
     * Reference to the StalessSessionBean that implements Purge function.
     */
    private static JasmineEventPurgeTaskSLBRemote purgeEjb = null;

    /**
     * Supported format for date parameters: executionStartDate.
     */
    private final static String DATE_FORMAT = "yyyy/MM/dd-HH:mm";

    /**
     * Formatter for DATE_FORMAT.
     */
    private Format formatter;

    /**
     * Command arguments.
     */
    protected String[] arguments = null;

    /**
     * Name of the command.
     */
    protected String name = "purge";

    /**
     * List of options that should be parsed from the command line.
     */
    private Options options = null;

    /**
     * The command line
     */
    private CommandLine commandLine = null;

    /**
     * Operation is purge task creation.
     */
    private boolean createTask = false;

    /**
     * Operation is print the list of purge tasks.
     */
    private boolean listTasks = false;

    /**
     * Operation is purge a task remove.
     */
    private boolean removeTask = false;

    /**
     * Operation is pause a purge a task.
     */
    private boolean pauseTask = false;

    /**
     * Operation is start a paused purge a task.
     */
    private boolean startTask = false;

    /**
     * Id of the task to remove/start/pause
     */
    private long taskId;

    /**
     * maxAge option
     */
    private long maxEntryAge = -1;

    /**
     * Default value when maxEntryAge is used as criteria.
     */
    private int maxEntryNumber= -1;

    /**
     * Date corresponding to startDate option
     */
    private Date executionStartDate = null;

    /**
     * execPeriod option
     */
    private long executionInterval;

    /**
     * By default, execute immediately
     * set to false if startDate provided
     */
    private boolean immediateExecution = true;

    /**
     * By default, execute purge corresponding to all
     * the targets.
     */
    private boolean executeOnAllServers = true;

    /**
     * domainName option
     */
    private String domainName;

    /**
     * serverName option
     */
    private String serverName;

    /**
     * The constructor allows to initialize the purge command's options
     */
    public PurgeDataBase() {
        this.setOptions();
        formatter = new SimpleDateFormat(DATE_FORMAT);
    }

    /**
     * Define all the possible options for this command.
     */
    private void setOptions() {
        // contains the description of all the possible options
        options = new Options();

        // create an Option group for purge operation
        OptionGroup action = new OptionGroup();
        action.setRequired(true);
        // create the options:
        // - createTask
        Option option = getOption("createTask");
        action.addOption(option);
        // - listTasks
        option = getOption("listTasks");
        action.addOption(option);
        // - removeTask
        option = getOption("removeTask");
        action.addOption(option);
        // - pauseTask
        option = getOption("pauseTask");
        action.addOption(option);
        // - pauseTask
        option = getOption("startTask");
        action.addOption(option);
        options.addOptionGroup(action);

        // create an Option group for maxAge and maxNumber
        OptionGroup ageOrNumber = new OptionGroup();
        ageOrNumber.setRequired(false);
        // create the options: maxAge
        option = getOption("maxAge");
        ageOrNumber.addOption(option);
        // maxEntries
        option = getOption("maxNumber");
        ageOrNumber.addOption(option);
        options.addOptionGroup(ageOrNumber);

        // server name
        option = getOption("serverName");
        options.addOption(option);
        // domain name
        option = getOption("domainName");
        options.addOption(option);

        // start date
        option = getOption("startDate");
        options.addOption(option);

        // execution period
        option = getOption("execPeriod");
        options.addOption(option);

    }

    /**
     * Allows to create a given option.
     * @param optionName the name of the option
     * @return The created Option object
     */
    private Option getOption(String optionName) {
        Option option = null;
        if (optionName.equals("maxAge")) {
            option = new Option("maxAge", "maxAge", true, "Entries older then maxAge will be purged");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("max entry age");
        } else if (optionName.equals("maxNumber")) {
            option = new Option("maxNumber", "maxNumber", true, "Maximum number of entries that are not purged");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("max entry number");
        } else if (optionName.equals("serverName")) {
            option = new Option("serverName", "serverName", true, "The entries that will be purged are related to this server");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("server name");
        } else if (optionName.equals("domainName")) {
            option = new Option("domainName", "domainName", true, "The entries that will be purged are related to this domain");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("domain name");
        } else if (optionName.equals("startDate")) {
            option = new Option("startDate", "startDate", true, "Date for purge to start");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("start date");
        } else if (optionName.equals("execPeriod")) {
            option = new Option("execPeriod", "execPeriod", true, "Time interval for purge execution");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("execution period");
        } else if (optionName.equals("createTask")) {
            option = new Option("createTask", "createTask", false, "Create purge task");
            option.setRequired(false);
        } else if (optionName.equals("listTasks")) {
            option = new Option("listTasks", "listTasks", false, "List tasks");
            option.setRequired(false);
        } else if (optionName.equals("removeTask")) {
            option = new Option("removeTask", "removeTask", true, "Remove Purge task");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("taskId");
        } else if (optionName.equals("pauseTask")) {
            option = new Option("pauseTask", "pauseTask", true, "Pause purge task");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("taskId");
        } else if (optionName.equals("startTask")) {
            option = new Option("startTask", "startTask", true, "Start purge task");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("taskId");
        }
        return option;
    }

    /**
     * Get the command arguments from the command line.
     * @param args
     * @throws ParseException
     * @throws AlreadySelectedException
     * @throws java.text.ParseException
     */
    private void setArgs(final String[] args) throws ParseException
        , AlreadySelectedException, MissingOptionException
        , java.text.ParseException, IllegalArgumentException {

        BasicParser bp = new BasicParser();
        commandLine = bp.parse(options, args);
        if (commandLine.hasOption("createTask")) {
            this.createTask = true;
        }
        if (commandLine.hasOption("listTasks")) {
            this.listTasks = true;
        }
        if (commandLine.hasOption("removeTask")) {
            this.removeTask = true;
            String idString = commandLine.getOptionValue("removeTask");
            setTaskId(idString);
        }
        if (commandLine.hasOption("pauseTask")) {
            this.pauseTask = true;
            String idString = commandLine.getOptionValue("pauseTask");
            setTaskId(idString);
        }
        if (commandLine.hasOption("startTask")) {
            this.startTask = true;
            setTaskId(commandLine.getOptionValue("startTask"));
        }
        if (commandLine.hasOption("maxAge")) {
            setMaxEntryAge(commandLine.getOptionValue("maxAge"));
        }
        if (commandLine.hasOption("maxNumber")) {
            setMaxEntryNumber(commandLine.getOptionValue("maxNumber"));
        }
        if (commandLine.hasOption("startDate")) {
            setExecutionStartDate(commandLine.getOptionValue("startDate"));
            immediateExecution = false;
        }
        if (commandLine.hasOption("execPeriod")) {
            setExecutionInterval(commandLine.getOptionValue("execPeriod"));
            immediateExecution = false;
        }
        if (commandLine.hasOption("serverName")) {
            setServerName(commandLine.getOptionValue("serverName"));
            executeOnAllServers = false;
        }
        if (commandLine.hasOption("domainName")) {
            setDomainName(commandLine.getOptionValue("domainName"));
            executeOnAllServers = false;
        }

    }

    /**
     * Command execution: call the sole business method defined by the purgeEjb, configureTimer()
     */
    private void exec() throws IllegalArgumentException {
        if (isCreateTask()) {
            createPurgeTask();
        } else if (isListTasks()) {
            listTasks();
        } else if (isRemoveTask()) {
            removeTask();
        } else if (isPauseTask()) {
            pauseTask();
        } else if (isStartTask()) {
            startTask();
        }
    }

    /**
     * Create a purge task based on the command options.
     * Print out the task id.
     */
    private void createPurgeTask() throws IllegalArgumentException {
        JasmineEventPurgeTask purgeTaskInitial = initPurgeTask();
        JasmineEventPurgeTask purgeTask = null;
        purgeTask = purgeEjb.configurePurgeTask(purgeTaskInitial);
        long purgeTaskId = purgeTask.getId();
        System.out.println("\nCreated purge task:");
        System.out.println(putgeTaskInfo(purgeTask));
    }

    /**
     * List all the tasks with the corresponding parameters
     * (corresponding creation options).
     */
    private void listTasks() {
        List<JasmineEventPurgeTask> purgeTasks = purgeEjb.listPurgeTasks();
        System.out.println("\nPurge tasks:");
        for (JasmineEventPurgeTask purgeTask : purgeTasks) {
            System.out.println(putgeTaskInfo(purgeTask));
        }

    }

    /**
     * Remove a task having the id given by the removeTask option value
     */
    private void removeTask() {
        long removedTaskId = purgeEjb.removePurgeTask(taskId);
        if (removedTaskId > 0) {
            System.out.println("\nRemoved purge task " + taskId);
        } else {
            System.out.println("\nTask " + taskId + " cannot be removed, does not exists !");
        }
    }

    /**
     * Pause a task having the id given by the pauseTask option value
     */
    private void pauseTask() {
        long purgedTaskId = purgeEjb.pausePurgeTask(taskId);
        if (purgedTaskId > 0) {
            System.out.println("\nPaused task " + taskId);
        } else {
            System.out.println("\nTask " + taskId + " cannot be paused, does not exists !");
        }
    }

    /**
     * Start a task having the id given by the startTask option value
     */
    private void startTask() {
        long startedTaskId = purgeEjb.startPurgeTask(taskId);
        if (startedTaskId > 0) {
            System.out.println("\nStarted task " + taskId);
        } else {
            System.out.println("\nTask " + taskId + " cannot be started, does not exists !");
        }
    }

    /**
     * Create a String that presents information about a purge task in a form
     * that corresponds to the DataManagement module presentation format.
     * @param purgeTask task id
     * @return All information about the given task.
     */
    private String putgeTaskInfo(JasmineEventPurgeTask purgeTask) {
        String info = new String();
        String id = "Id " + purgeTask.getId();
        boolean isStarted = purgeTask.isStarted();
        boolean isAllServer = purgeTask.isExecuteOnAllServers();
        String domainName = purgeTask.getDomainName();
        String serverName = purgeTask.getServerName();
        long maxAgeValue = purgeTask.getMaxEntryAge();
        long maxNumberValue = purgeTask.getMaxEntryNumber();
        String maxAge;
        if (maxAgeValue == 0) {
            maxAge = "; Max Age -";
        } else {
            maxAge = "; Max Age " + new Long(maxAgeValue).toString();
        }
        String maxNumber;
        if (maxNumberValue == 0) {
            maxNumber = "; Max Number -";
        } else {
            maxNumber = "; Max Number " + new Long(maxNumberValue).toString();
        }
        String startDate = "; Start Date " + purgeTask.getExecutionStartDate();
        String period = "; Period " + purgeTask.getExecutionInterval();
        Date lastExecutionDate = purgeTask.getLastExecutionDate();
        String lastExecutionString;
        if (lastExecutionDate != null) {
            lastExecutionString = lastExecutionDate.toString();
        } else {
            lastExecutionString = " ";
        }
        String lastExecution = "; Last Execution " + lastExecutionString + " : " + purgeTask.getRemovedEvents() + " events";
        info = info + id;
        if (isStarted) {
            info = info + "; STARTED";
        } else {
            if (purgeTask.getExecutionInterval() != 0) {
                info = info + "; PAUSED ";
            } else {
                info = info + ";        ";
            }
        }
        if (isAllServer) {
            info = info + "; all";
        } else {
            info = info + "; " + serverName + "/" + domainName;
        }
        info = info + startDate + period + maxAge + maxNumber + lastExecution;
        return info;
    }

    /**
     * Construct the purgeEjb's configureTimer() method's argument.
     * @return
     */
    private JasmineEventPurgeTask initPurgeTask() throws IllegalArgumentException {

        if (!validOptions()) {
            throw new IllegalArgumentException("Invalid purge criteria options.");
        }

        JasmineEventPurgeTask event = new JasmineEventPurgeTask();

        event.setMaxEntryAge(getMaxEntryAge());
        event.setMaxEntryNumber(getMaxEntryNumber());
        event.setExecutionStartDate(getExecutionStartDate());
        event.setExecutionInterval(getExecutionInterval());
        event.setExecuteOnAllServers(isExecuteOnAllServers());
        event.setDomainName(getDomainName());
        event.setServerName(getServerName());

        return event;
    }

    /**
     * @return true if purge task create options are valid.
     */
    private boolean validOptions() {
        boolean valid = true;
        // check that either maxAge or maxNumber provided
        if (getMaxEntryNumber() < 0 && getMaxEntryAge() < 0) {
            valid = false;
        }
        return valid;
    }

    /**
     * @return true if create purge task operation is executed
     */
    private boolean isCreateTask() {
        return createTask;
    }

    /**
     * @return true if list purge tasks operation is executed
     */
    private boolean isListTasks() {
        return listTasks;
    }

    /**
     * @return true if remove purge task operation is executed
     */
    public boolean isRemoveTask() {
        return removeTask;
    }
    /**
     * @return true if pause purge task operation is executed
     */
    public boolean isPauseTask() {
        return pauseTask;
    }

    /**
     * @return true if start purge task operation is executed
     */
    public boolean isStartTask() {
        return startTask;
    }

    public long getTaskId() {
        return taskId;
    }

    public void setTaskId(String taskId) {
        this.taskId = (new Long(taskId)).longValue();
    }

    /**
     * @return maxEntryAge the value of maxEntryAge as set by the -maxAge option
     */
    private long getMaxEntryAge() {
        return maxEntryAge;
    }

    /**
     * Determine maxEntryAge in milliseconds from a String having DURATION_FORMAT.
     * @param maxEntryAge_param -maxAge option
     */
    private void setMaxEntryAge(String maxEntryAge_param) throws IllegalArgumentException {
        maxEntryAge = getMilliSeconds(maxEntryAge_param);
    }

    /**
     *
     * @return maxEntryNumber value as set by -maxNumber option
     */
    private int getMaxEntryNumber() {
        return maxEntryNumber;
    }

    /**
     * Determine maxEntryNumber
     * @param maxEntryNumber_param the -maxNumber option
     */
    private void setMaxEntryNumber(String maxEntryNumber_param) {
        maxEntryNumber = (new Integer(maxEntryNumber_param)).intValue();
    }

    /**
     * @return the value of executionStartDate as set by the -startDate option
     */
    private Date getExecutionStartDate() {
        return executionStartDate;
    }

    /**
     * Determine executionStartDate from a String having the DATE_FORMAT
     * @param executionStartDate_param the -startDate option
     * @throws java.text.ParseException
     */
    private void setExecutionStartDate(String executionStartDate_param) throws java.text.ParseException {
        Date date = null;
        try {
            date = (Date) formatter.parseObject(executionStartDate_param);
        } catch (java.text.ParseException e) {
            throw e;
        }
        executionStartDate = date;
        System.out.println("Periodic execution starts : " + executionStartDate);
    }

    /**
     * Default start date for periodic execution.
     * This means this is a periodic execution.
     * TODO Remove this method ?
     */
    private void setExecutionStartDate() {
        executionStartDate = Calendar.getInstance().getTime();
        System.out.println("Periodic execution starts now : " + executionStartDate);
    }

    private long getExecutionInterval() {
        return executionInterval;
    }


    private void setExecutionInterval(String executionInterval_param) throws IllegalArgumentException {
        executionInterval = getMilliSeconds(executionInterval_param);
    }

    private boolean isExecuteOnAllServers() {
        return executeOnAllServers;
    }

    private boolean isImmediateExecution() {
        return immediateExecution;
    }

    private String getServerName() {
        return serverName;
    }

    private void setServerName(String serverName_param) {
        serverName =  serverName_param;
    }

    private String getDomainName() {
        return domainName;
    }

    private void setDomainName(String domainName_param) {
        domainName = domainName_param;
    }

    /**
     *
     * @param duration String having DURATION_FORMAT.
     * @return milliseconds corresponding to the number of days, hours and minutes.
     * @throws IllegalArgumentException The format is not respected.
     * For example, the days, hours and minutes have to be provided in this order.
     */
    private long getMilliSeconds(String duration) throws IllegalArgumentException {
        duration = duration.toLowerCase();
        String daysMarquer = "d";
        String hoursMarquer = "h";
        String minutesMarquer = "m";
        long numberOfMilliSeconds = 0;
        int numberOfDays = 0;
        int numberOfHours = 0;
        int numberOfMinutes = 0;
        int dayIndex = duration.indexOf(daysMarquer);
        int hoursIndex = duration.indexOf(hoursMarquer);
        int minutesIndex = duration.indexOf(minutesMarquer);
        if (dayIndex == -1 && hoursIndex == -1 && minutesIndex == -1) {
            throw new IllegalArgumentException("Duration " + duration + " is missformed. The correct form is daysDhoursHminutesM");
        }
        // check that order is respected : days - hours - minutes
        boolean orderOk = true;
        if (minutesIndex > 0) {
            if (minutesIndex < dayIndex || minutesIndex < hoursIndex) {
                orderOk = false;
            }
        }
        if (hoursIndex > 0) {
            if (hoursIndex < dayIndex) {
                orderOk = false;
            }
        }
        if (!orderOk) {
            throw new IllegalArgumentException("Duration " + duration + " is missformed. The correct form is daysDhoursHminutesM");
        }
        // look for number of days
        int index = duration.indexOf(daysMarquer);
        if (index > 0) {
            int beginIndex = 0;
            String daysValue = duration.substring(beginIndex, index);
            duration = duration.substring(index + 1, duration.length());
            numberOfDays = new Integer(daysValue) ;
        }
        numberOfHours = 24 * numberOfDays;

        // look for number of hours
        index = duration.indexOf(hoursMarquer);
        if (index > 0) {
            int beginIndex = 0;
            String hoursValue = duration.substring(beginIndex, index);
            duration = duration.substring(index + 1, duration.length());
            numberOfHours = numberOfHours + new Integer(hoursValue);
        }
        numberOfMinutes = 60 * numberOfHours;

        // look for number of minutes
        index = duration.indexOf(minutesMarquer);
        if (index > 0) {
            int beginIndex = 0;
            String minutesValue = duration.substring(beginIndex, index);
            numberOfMinutes = numberOfMinutes + new Integer(minutesValue);
        }
        numberOfMilliSeconds = 60000 * numberOfMinutes;
        return numberOfMilliSeconds;

    }

    /**
     *
     * @throws NamingException
     */
    private static void initEJBConnection() throws NamingException {
        Hashtable<String, Object> env = new Hashtable<String, Object>();
        env.put(Context.INITIAL_CONTEXT_FACTORY, SmartContextFactory.class.getName());
        Context context = new InitialContext(env);
        purgeEjb = (JasmineEventPurgeTaskSLBRemote) context.lookup(EJB_JNDI_NAME);
    }


      /**
     * The main
     *
     * @param args command line arguments
     */
    public static void main(final String[] args) {
        PurgeDataBase purge = null;
        try {
            purge = new PurgeDataBase();

            if (args.length <= 0) {
                purge.help();
                System.exit(-1);
            }

            purge.setArgs(args);

            initEJBConnection();

            purge.exec();

        } catch (AlreadySelectedException ase) {
            System.err.println("Error in purge command arguments: " + ase.getMessage());
            purge.help();
        } catch (java.text.ParseException pe) {
            System.err.println("Error in purge command arguments: " + pe.getMessage());
            purge.help();
        } catch (MissingOptionException mi) {
            System.err.println("Error in purge command arguments: " + mi.getMessage());
            purge.help();
        } catch (NamingException ne) {
            System.err.println("Cannot connect to the JASMINe server: " + ne.getMessage());
            ne.printStackTrace();
        } catch (IllegalArgumentException ia) {
            System.err.println("Error in purge command arguments: " + ia.getMessage());
            purge.help();

        } catch (Exception e) {
            e.printStackTrace();
            System.exit(-1);
        }
    }

    public final String getName() {
        return this.name;
    }

    /**
     * Implementation of inherited method.
     *
     * Displays the help for this command.
     *
     * @see org.ow2.jasmine.monitoring.mbeancmd.Command#help()
     * @see AbstractCommand#getHelpDoc()
     */
    public void help() {
        ClassLoader classL = Thread.currentThread().getContextClassLoader();
        InputStream in0 = Thread.currentThread().getContextClassLoader().getResourceAsStream(this.getHelpDoc());

        if (in0 != null) {
            BufferedReader in = new BufferedReader(new InputStreamReader(in0));
            OutputStreamWriter out = new OutputStreamWriter(System.out);
            try {

                boolean goOn = true;
                char[] buf = new char[2048];
                while (goOn) {
                    int read;

                    read = in.read(buf);
                    if (read > 0) {
                        out.write(buf, 0, read);
                        out.flush();
                    } else {
                        goOn = false;
                        System.out.println();
                    }
                }

            } catch (IOException e) {
                logger.error("Error reading help ", e);
            } finally {

                if (in != null) {
                    try {
                        in.close();
                    } catch (IOException e) {
                        logger.error("Error closing the BufferedReader ", e);
                    }
                }
            }
        } else {
            System.out.println("No help available.");
        }

    }

    /**
     * Gets the help document corresponding to this command.
     *
     * @return Name of the help document corresponding to this command.
     */
    protected String getHelpDoc() {
        String classRadix = this.getClass().getName().substring(this.getClass().getName().lastIndexOf(".") + 1);
        String packName = this.getClass().getPackage().getName();
        String helpdoc = packName.substring(0, packName.lastIndexOf("."));
        helpdoc = helpdoc.replaceAll("[.]", "/") + "/helpdoc/" + classRadix + ".help";
        return helpdoc;
    }


    public String summary() {
        return "Purge JASMINe DataBase";
    }
}
